#ifndef SMALLSORT_H
#define SMALLSORT_H

#ifndef SMALL
#define SMALL 64
#endif
#ifndef BUFFER_SIZE
#define BUFFER_SIZE 500
#endif
#ifndef BASE
#define BASE buffered
#endif
#ifndef SORTER
#define SORTER smallSort
#endif
#ifndef SSMALL
#define SSMALL 2048
#endif
#ifndef SBASE
#define SBASE smallSort
#endif

//#define PRINT_RESULTS

#include <iostream>
#include <iomanip>
#include <algorithm>
#include <map>

using namespace std;

#define Assert1(a) void(0)

// From SSSS-Paper Sanders

// customized basic sorting routines
template <class T> inline void 
sort2(T& a, T& b) { if (b < a) swap(a, b); }

template <class T> inline void 
sort3(T& a, T& b, T& c) {
  T temp;
  if (b < a) { 
    if (c < a) {                         // b , c < a
      if (b < c) {                       // b < c < a
        temp = a; a=b;  b=c;  c=temp;
      } else {                           // c <=b < a
        swap(c,a);
      }
    } else {                             // b < a <=c
      swap(a, b);
    }
  } else {                               // a <=b
    if (c < a) {                         // c < a <=b
      temp=c; c=b; b=a;  a=temp;
    } else {                             // a <=b , c
      if (c < b) {                       // a <=c < b
        swap(b,c);
      }
    }
  }
  Assert1(!(b < a) && !(c < b));
}


template <class T> inline void
sort4(T& a, T& b, T& c, T& d) {
  sort2(a, b);
  sort2(c, d);    // abcd acbd acdb cabd cadb cdab
  if (a < c) {    // abcd acbd acdb
    if (c < b) {  //      acbd acdb
      if (b < d) {//      acbd
        swap(b, c);
      } else     {//           acdb
        T t = b; b = c; c = d; d = t;
      } 
    } // else        abcd
  } else {        //                cabd cadb cdab
    if (a < d) {  //                cabd cadb
      if (b < d) {//                cabd
        T t = a; a = c; c = b; b = t;
      } else {    //                     cadb
        T t = a; a = c; c = d; d = b; b = t;
      }
    } else {      //                          cdab
      swap(a,c);
      swap(b,d);
    }
  }
  Assert1(!(b < a) && !(c < b) & !(d < c));
}

//not used
template <class T> inline void
sort5(T& a, T& b, T& c, T& d, T& e) {
  sort2(a, b);
  sort2(d, e);
  if (d < a){
    swap(a, d); 
    swap(b, e); 
  }                   // a < d < e, a < b
  if (d < c) {
    swap(c,d);        // a minimal, c < {de}
    sort2(d,e);
  } else {            // a<b, a<d<e, c<d<e
    sort2(a, c);
  }                   // a min, c < d < e
  // insert b int cde by binary search
  if (b < d) {        // {bc} < d < c
    sort2(b, c);
  } else {            // c < d < {bc}
    T oldB = b; 
    T oldC = c; 
    b = c;  c = d;
    if (oldB < oldC) { d = oldB;  e = oldC; }
    else             { d = oldC;  e = oldB; }
  }
  Assert1(!(b < a) && !(c < b) & !(d < c) & !(e < d));
}

template <class T> inline void
insertionSort(T* a, T* aEnd) {
  T* pp;
  for (T* p=a+1;  p < aEnd;  p++) {
    // Invariant a..p-1 is sorted;
    T t = *p;
    if (t < *a) { // new minimum
      // move stuff to the right
      for (pp = p; pp != a;  pp--) {
        *pp = *(pp-1);
      }
      *pp = t;
    } else {
      // now we can use *a as a sentinel
      for (pp = p;  t < *(pp-1);  pp--) {
        *pp = *(pp-1);
      }  
      *pp = t;
    }
  }
}

template <class T> inline void
insertionSort2(T* a, T* aEnd) {
  // Only one number wrong, probably because this has to be even, fix me later
  T* pp;

  T* min;
  T* max;

  for (T* p=a+(aEnd-a+1)%2;  p < aEnd;  p++) {
    // Invariant a..p-1 is sorted;
    //T t = *p;
    if (*p > *(p+1)) {
      swap(*p,*(p+1));
    }

    min = *p;
    max = *(p+1);

    if (max < *a) { // new minimum
      // move stuff to the right
      for (pp = p-1; pp != a;  pp--) {
        *(pp+2) = *pp;
      }
    } else {
      for (pp = p-1; max < *pp;  pp--) {
        *(pp+2) = *pp;
      }
    }

    *(pp+2) = max;

    //if (min < *a) { // new minimum
    //  // move stuff to the right
    //  for (; pp != a;  pp--) {
    //    *(pp+1) = *pp;
    //  }
    //} else {
      // now we can use *a as a sentinel
      for (; min < *pp;  pp--) {
        *(pp+1) = *pp;
      }
    //}

    *(pp+1) = min;
  }
}

// sort a small number of elements
template <class T>
inline void smallSort(T* a, T* b) {
  int n = b - a;
  switch (n) {
  case 0:
  case 1: break;
  case 2: sort2(a[0], a[1]); break;
  case 3: sort3(a[0], a[1], a[2]); break;
  case 4: sort4(a[0], a[1], a[2], a[3]); break;
  default: 
    //Debug3(cout << "sort(" << n << ")" << endl);
    //if (n <= 100)
      insertionSort(a, b);
    //else {
    //  std::sort(a, a+n);
    //  //make_heap(a, a + n);
    //  //sort_heap(a, a + n);
    //}
  }
}

template <class T> void mine2(T* a, T* b)
{
  if (b-a == 1) {
  } else if (b-a == 2) {
    sort2(a[0], a[1]);
  } else if (b-a == 3) {
    sort3(a[0], a[1], a[2]);
  } else if (b-a == 4) {
    sort4(a[0], a[1], a[2], a[3]);
  } else if (b-a == 5) {
    sort5(a[0], a[1], a[2], a[3], a[4]);
  //} else if (b-a <= 100) {
  } else {
    insertionSort(a, b);
  //} else {
  //  std::sort(a, b);
  }
}

template <class T> void mine(T* a, T* b)
{
  if (b-a == 1) {
  } else if (b-a == 2) {
    sort2(a[0], a[1]);
  } else if (b-a == 3) {
    sort3(a[0], a[1], a[2]);
  } else if (b-a == 4) {
    sort4(a[0], a[1], a[2], a[3]);
  } else if (b-a == 5) {
    sort5(a[0], a[1], a[2], a[3], a[4]);
  } else if (b-a <= 100) {
    buffered(a, b);
  } else {
    std::sort(a, b);
  }
}

#define CMP_SWAP(i, j) if (a[i] > a[j]) \
  { T tmp = a[i]; a[i] = a[j]; a[j] = tmp; }

template <class T> void sort8_network(T* a)
{
    CMP_SWAP(0, 1); CMP_SWAP(2, 3); CMP_SWAP(4, 5); CMP_SWAP(6, 7);
    CMP_SWAP(0, 2); CMP_SWAP(1, 3); CMP_SWAP(4, 6); CMP_SWAP(5, 7);
    CMP_SWAP(1, 2); CMP_SWAP(5, 6); CMP_SWAP(0, 4); CMP_SWAP(1, 5);
    CMP_SWAP(2, 6); CMP_SWAP(3, 7); CMP_SWAP(2, 4); CMP_SWAP(3, 5);
    CMP_SWAP(1, 2); CMP_SWAP(3, 4); CMP_SWAP(5, 6);
}

template <class T> void baseCase(T* l, T* r) {
  sort(l, r);
}

template <class T> void sortInsertion(T* l, T* r)
{
  T tmp;
  T* j;
  for (T* i = l; i < r; i++)
  {
    tmp = *i;
    j = i;
    for (; j > l && tmp < *(j - 1); --j)
        *j = *(j - 1);
    *j = tmp;
  }
}

template <class T> void sort2Insertion(T* l, T* r)
{
  T min;
  T max;
  T* i;
  T* j;

  // Still one number wrong
  for (i = l + (r-l+1)%2; i < r; i++) {
    if (*i <= *(i+1)) {
      min = *i;
      max = *(i+1);
    } else {
      max = *i;
      min = *(i+1);
    }
    j = i - 1;
    while (*j > max) {
      *(j+2) = *j;
      j--;
    }
    *(j+2) = max;
    while (*j > min) {
      *(j+1) = *j;
      j--;
    }
    *(j+1) = min;
  }
}

template <class T> void sort4Insertion(T* l, T* r)
{
  T m1, m2, m3, m4;
  T* i;
  T* j;

  // No idea if this works
  for (i = l + (r-l+1)%4; i < r; i++) {
    m1 = *i;
    m2 = *(i+1);
    m3 = *(i+2);
    m4 = *(i+3);

    sort4(m1, m2, m3, m4);

    j = i - 1;
    while (*j > m4) {
      *(j+4) = *j;
      j--;
    }
    *(j+4) = m4;
    while (*j > m3) {
      *(j+3) = *j;
      j--;
    }
    *(j+3) = m3;
    while (*j > m2) {
      *(j+2) = *j;
      j--;
    }
    *(j+2) = m2;
    while (*j > m1) {
      *(j+1) = *j;
      j--;
    }
    *(j+1) = m1;
  }
}

struct x {
  int* l;
  int* r;
};


const int MAXN = 4*1024*1024;
//int data[MAXN];

int inc[1000];

inline int randomBin(int n) 
{
  //int result = (int)(genrand() * n);
  int result = (int)(drand48() * n);
  //Assert2(result >= 0);
  //Assert2(result <  n);
  return result;
}

#define MAXVAL (1<<30)

// find k about equally spaced increments
double avgFactor;
int initInc(int k)
{
  int p;
  double current = 1.0;
  inc[0] = int(current);
  for (p=1;  p < k;  p++) {
    current *= avgFactor;
    int lower = int(0.75*current) + 1;
    int upper = int(current) + 1;
    inc[p] = lower + randomBin(upper - lower);
    // inc[p] = inc[p-1]*avgFactor + 1;
    //Debug0(cout << upper << " " << lower << " " << inc[p] << endl);
  }
  return p; // use that many phases and increments inc[0..p-1]
}

int KK;

// shell sort using the pseudocode from Duden Informatik
template <class T> void shellsort(T* l, T* r)
{  
  //int count = 0;
  int s; // phase counter

  for (s = KK;s >= 0;  s--) { // phase s
    int incr = inc[s];
    //Debug3(cout << "incr="<<incr <<" "; printData(n));
    for (int j = incr;  j < r - l;  j++) {
      int i = j-incr;
      int x = l[j];
      while (i >= 0) {
        //count++;
        if (x >= l[i]) break;
        l[i+incr] = l[i];
        i -= incr;
      }
      l[i + incr] = x;
    }
  }
  //return count;
}

bool operator< (const x &cC1, const x &cC2)
{
    return cC1.l < cC2.l;
}

int* start;
int* pend;

template <class T> void newbuffered(T* newl, T* newr)
{
  if (!start) {
    start = newl;
  }
  pend = newr;

  if (newr - start > 100) {
    newemptyBuffer();
  }
}

void newemptyBuffer()
{
  if (start && pend)
    SORTER(start, pend);
  start = NULL;
}

#ifdef ONE_BUFFER
x buffer[BUFFER_SIZE];
#else
int* lbuffer[BUFFER_SIZE];
int* rbuffer[BUFFER_SIZE];
#endif
int bufferPos = 0;

template <class T> void buffered(T* newl, T* newr)
{
  if (bufferPos < BUFFER_SIZE) {
#ifdef ONE_BUFFER
    buffer[bufferPos] = x{newl, newr};
#else
    lbuffer[bufferPos] = newl;
    rbuffer[bufferPos] = newr;
#endif
    bufferPos++;
  } else {
    emptyBuffer();
    //if (newr - newl > 4 * SMALL) {
    //  quickSort(newl, newr-1);
    //} else {
      SORTER(newl, newr);
    //}
  }
}

void emptyBuffer()
{
#ifdef BUFFER_COMBINE
  int* l;
#endif
  int* r;
  int y;

#ifdef BUFFER_COMBINE
#ifdef BUFFER_SORT
  sort(buffer, buffer + BUFFER_SIZE);
#endif
#endif
#ifdef BUFFER_REVERSE
  for (int x = bufferPos - 1; x >= 0; x--) {
#else
  for (int x = 0; x < bufferPos; x++) {
#endif
#ifdef BUFFER_COMBINE
#ifdef ONE_BUFFER
    l = buffer[x].l;

#ifdef BUFFER_REVERSE
    while (x >= 0 && buffer[x].r == buffer[x-1].l) {
      x--;
    }
#else
    while (x < BUFFER_SIZE && buffer[x].r == buffer[x+1].l) {
      x++;
    }
#endif
    SORTER(l, buffer[x].r);
#else
    l = lbuffer[x];

#ifdef BUFFER_REVERSE
    while (x >= 0 && rbuffer[x] == lbuffer[x-1]) {
      x--;
    }
#else
    while (x < BUFFER_SIZE && rbuffer[x] == lbuffer[x+1]) {
      x++;
    }
#endif
    SORTER(l, rbuffer[x]);
#endif
#else
#ifdef ONE_BUFFER
    SORTER(buffer[x].l, buffer[x].r);
#else
    SORTER(lbuffer[x], rbuffer[x]);
#endif
#endif
  }
  bufferPos = 0;
}

template <class T> void nothing(T* l, T* r) {
}

#endif
