#include <iostream>
#include <iomanip>
#include <algorithm>
#include <map>

#include "smallsort.h"
#include "measuretime.h"
#include "util.h"
#include "log2.h"
#include "findsamples.h"

using namespace std;

map<string, double> times;
#define TIME_BEGIN(timeVar) Timer timer_##timeVar
#define TIME_END(timeVar) times[#timeVar] += timer_##timeVar.elapsed()

//#define DETAILED_TIME_MEASUREMENT
#ifdef DETAILED_TIME_MEASUREMENT

#define MEASURETIME_BEGIN(timeVar) Timer timer_##timeVar
#define MEASURETIME_END(timeVar) times[#timeVar] += timer_##timeVar.elapsed()

#else
#define MEASURETIME_BEGIN(timeVar) void(0)
#define MEASURETIME_END(timeVar) void(0)
#endif

#define PRINT(a, n) \
do { \
    cout << __FUNCTION__ << " -> " << #a << " = {"; \
    for (int i = 0; i < n; ++i) \
    { \
        if(i) cout << ", "; \
        cout << a[i]; \
    } \
    cout << "}" << endl; \
} while(false)

#undef PRINT
#define PRINT(a, n) void(0)

template <class T> void qsort2(T* l, T* r) {
  quickSort(l, r-1);
}

template <class T> inline T* badPivot(T* l, T* middle, T* r) {
  if (*l > *middle) {
    if (*l > *r) {
      return l;
    } else if (*r > *middle) {
      return r;
    } else {
      return middle;
    }
  } else if (*r > *middle) {
    return r;
  } else {
    return middle;
  }
}

template <class T> void quickSort(T* l, T* r) {
  size_t size;
  while (r - l > SMALL) {
    size = r - l;
    T* middle = l + size / 2;

    T* j = middle; // Pivot = middle
    //T* j = l + 10 * size / 11;
    //T* j = badPivot(l, middle, r);
    //T* j = max_element(l, 10 < size ? l + 10 : r);
    swap(*l, *j);
    T p = *l;
    T* i = l;
    j = r;

    do {
      while (*i < p) i++;
      while (*j > p) j--;
      if (i <= j) {
        swap(*i, *j);
        i++;
        j--;
      }
    } while (i <= j);

    if (i < middle) {
      quickSort(l, j);
      l = i;
    } else {
      quickSort(i, r);
      r = j;
    }
  }

  BASE(l, r+1);

  //if (r - l == SMALL) {
  //  MEASURETIME_BEGIN(base8);
  //  //sortInsertion(l, r + 1);
  //  sort2Insertion(l, r + 1);
  //  //sort8_network(l);
  //  MEASURETIME_END(base8);
  //} else {
    //MEASURETIME_BEGIN(base);
    //baseCase(l, r + 1);
    //sortInsertion(l, r + 1);
    //insertionSort(l, r + 1);
    //buffered(l, r + 1);
    //sort2Insertion(l, r + 1);
    //sort8_network(l);
    //smallSort(l, r - l + 1);
    //#endif
    //B(l, r + 1);
    //MEASURETIME_END(base);
  //}
}

// Sample Sort by Misch Sadler, Sebastian Lehmann, Eric Braun

int *oracle;

#ifndef NUM_BUCKETS
#define NUM_BUCKETS 64
#endif
#ifndef SAMPLE_FACTOR
#define SAMPLE_FACTOR 2
#endif

const int NUM_SAMPLES = NUM_BUCKETS * SAMPLE_FACTOR - 1;

template <class T, int k = NUM_BUCKETS>
void findOraclesAndCount(T* const a, const int n, const T* const s, int* const oracle, int* const bucketCounter)
{
    PRINT(a, n);
    for (int i = 0; i < n; i++)
    {
        int j = 1;
        for(int x = 0; x < LOG2<k>::value; ++x) {
            j = j*2 + (a[i] > s[j]);
        }
        int b = j-k;
        bucketCounter[b]++;
        oracle[i] = b;
    }
    PRINT(oracle, n);
    PRINT(bucketCounter, k);
}

template <class T, int k = NUM_BUCKETS, class FindSamplesFunc>
void sampleSort(T* const a, const int n, FindSamplesFunc findSamplesFunc)
{
    if (n <= SSMALL)
    {
        MEASURETIME_BEGIN(baseCase);
        SBASE(a, a + n);
#if (BASE == nothing)
        insertionSort(a, a+n);
#endif
#if (BASE == buffered)
        emptyBuffer();
#endif
#if (BASE == newbuffered)
        newemptyBuffer();
#endif
        MEASURETIME_END(baseCase);
        return;
    }

    MEASURETIME_BEGIN(sample);
    T s[k];
    findSamplesFunc(a+0, a+n, s+1);  //s[0] is ignored
    MEASURETIME_END(sample);

    // Use base case sorting if samples are equal (first is not less than last)
    if(!(s[1] < s[NUM_BUCKETS - 1]))
    {
        MEASURETIME_BEGIN(baseCase);
        SBASE(a, a + n);
        MEASURETIME_END(baseCase);
        return;
    }


    MEASURETIME_BEGIN(findBuckets);
    // Find target buckets and counts
    int bucketCounter[k];
    for(int i = 0; i < k; ++i)
        bucketCounter[i] = 0;
    findOraclesAndCount(a, n, s, oracle, bucketCounter);
    MEASURETIME_END(findBuckets);

    // Calculate bucket offsets
    MEASURETIME_BEGIN(findBucketOffsets);
    int bucketOffset[k];
    int currentOffset = bucketOffset[0] = 0;
    for(int i = 1; i < k; ++i)
        bucketOffset[i] = (currentOffset += bucketCounter[i - 1]);

    int bucketPos[k];
    for(int i = 0; i < k; ++i)
        bucketPos[i] = bucketOffset[i];
    MEASURETIME_END(findBucketOffsets);


    MEASURETIME_BEGIN(distribute);
    // Bucket sort
    T* buckets = new T[n];
    for(int i = 0; i < n; ++i)
    {
        int b = oracle[i];
        buckets[bucketPos[b]++] = a[i];
    }
    for(int i = 0; i < n; ++i)
        a[i] = buckets[i];
    PRINT(buckets, n);
    delete[] buckets;
    MEASURETIME_END(distribute);

    // Recursive call
    for(int i = 0; i < k; ++i)
        if(bucketCounter[i])
            sampleSort(a + bucketOffset[i], bucketCounter[i], findSamplesFunc);
}

template <class T> void printArray(T* a, int n, char *comment) {
  for (int i = 0;  i < n;  i++) {
    cout << a[i] << endl ;
  }
}

int main(int argc, char** argv) {
  int n = atoi(argv[1]);
  int repeat = atoi(argv[2]);
  int mult = atoi(argv[3]);
  //int seed = atoi(argv[3]);
  //sgenrand(seed);
  srand(0);

  int *a = new int[n];

  oracle = new int[n];

  start = NULL;
  pend = NULL;

  for (int r = 0; r < repeat; r++) {
    for (int i = 0; i < n;  i++) {
      //a[i] = int(1e9*genrand_double());
      a[i] = rand();
    }

    //printArray(a, n, "input");

    TIME_BEGIN(total);
#ifdef STDSORT
    sort(a, a + n);
#elif SSSS
    sampleSort(a, n, findSamplesMultiRandomStdSort<int, NUM_BUCKETS, SAMPLE_FACTOR, int*>);
#else
    quickSort(a, a + n - 1);
#endif
#ifdef SORT_AFTERWARDS
    SORTER(a, a + n); // Can also run this instead of doing basecases, works best with SMALL 64
#endif
#if (BASE == buffered)
    emptyBuffer();
#endif
#if (BASE == buffered)
    newemptyBuffer();
#endif
    TIME_END(total);
  }

#ifdef PRINT_RESULTS
  printArray(a, n, "sorted");
#endif

  for(auto j = times.begin(); j != times.end(); ++j)
  {
    double time = j->second;
    cout << setw(8) << right << n << setw(12) << right << (long long)(time * mult) << endl;
  }
}
